from typing import Any, Dict, Tuple

import re

from .SchemaPrimitive import SchemaPrimitive


class SchemaPattern(SchemaPrimitive):
    """
    A schema requiring a match for a regex pattern.
    """

    pattern: str = None

    def __init__(self, pattern: str) -> None:
        super().__init__(python_type=str)
        self.pattern = pattern

    def __call__(
            self,
            value: str,
            dom_info: Tuple = None
    ) -> Any:
        if not re.match(self.pattern, value):
            raise ValueError(f"Parameter value {value} does not match regex pattern {self.pattern}.")
        return super().__call__(value)

    @property
    def jsonschema_definition(self) -> Dict[str, Any]:
        return {
            "type": self.type_name,
            "pattern": self.pattern
        }



from .Schema import Schema
from .SchemaAnything import SchemaAnything
from .SchemaType import SchemaType
from .SchemaPrimitive import SchemaPrimitive
from .SchemaNone import SchemaNone
from .SchemaConst import SchemaConst
from .SchemaEnum import SchemaEnum
from .SchemaPattern import SchemaPattern



from typing import Dict

from ..base_schema import Schema
from .UserObject import UserObject


def properties(user_object: UserObject) -> Dict[str, Schema]:
    """
    Retrieve the property names and their associated :class:`~wysdom.base_schema.Schema`
    objects for a particular `:class:`~wysdom.user_objects.UserObject.

    :param user_object: A `:class:`~wysdom.user_objects.UserObject
    :return:            A dictionary mapping property names to their :class:`~wysdom.base_schema.Schema`
    """
    return user_object.__json_schema_properties__.properties



from typing import Type, Any, Union, Optional, Callable

from ..base_schema import Schema
from ..object_schema import SchemaDict

from .UserProperty import UserProperty


class DictProperty(UserProperty):
    """
    A data descriptor for creating attributes in user-defined subclasses
    of :class:`~.wysdom.user_objects.UserObject` with a property_type
    of :class:`~.wysdom.object_schema.SchemaDict`.

    :param items:            The permitted data type or schema for the properties of the underlying
                             :class:`~.wysdom.object_schema.SchemaDict`. Must be one of:

                             A primitive Python type (str, int, bool, float)
                             A subclass of :class:`~.wysdom.user_objects.UserObject`.
                             An instance of :class:`~.wysdom.base_schema.Schema`.

    :param optional:         Determines whether this property is optional in the underlying
                             data object. If default or default_function are set, this
                             will default to True, otherwise False.

    :param name:             The name of this property in the underlying
                             data object. If not provided, this defaults to
                             the name of the attribute on the :class:`~.wysdom.user_objects.UserObject`
                             instance that owns the property.

    :param default:          A static value which provides a default value
                             for this property. Cannot be set in conjunction
                             with `default_function`.

    :param default_function: A function which provides a default value
                             for this property. The function must have a
                             single positional argument, `self`, which is
                             passed the :class:`~.wysdom.user_objects.UserObject` instance that
                             owns the property. Cannot be set in conjunction
                             with `default`.

    :param persist_defaults: If this property is set to True and a UserProperty has either the
                             `default` or `default_function` property, when the UserProperty returns
                             a default value that value will also be explicitly stored in the underlying
                             data object. This is often desirable behavior if the UserProperty
                             returns another object and your code expects it to return the same
                             object instance each time it is accessed.

    :param key_pattern:      A regex pattern to validate the keys of the dictionary against.
    """

    def __init__(
            self,
            items: Union[Type, Schema],
            optional: Optional[bool] = None,
            name: Optional[str] = None,
            default: Optional[Any] = None,
            default_function: Optional[Callable] = None,
            persist_defaults: Optional[bool] = None,
            key_pattern: Optional[str] = None
    ) -> None:
        super().__init__(
            property_type=SchemaDict(items, key_pattern=key_pattern),
            optional=optional,
            name=name,
            default=default,
            default_function=default_function,
            persist_defaults=persist_defaults
        )



from .UserObject import UserObject, UserProperties
from .UserProperty import UserProperty
from .ListProperty import ListProperty
from .DictProperty import DictProperty
from .functions import properties



from __future__ import annotations

from typing import Type, Union
from enum import Enum

import inspect

from ..base_schema import Schema
from ..dom import DOMElement

from ..base_schema import SchemaPrimitive, SchemaEnum


def resolve_arg_to_schema(
        arg: Union[Type, Schema]
) -> Schema:
    """
    Resolve an argument of heterogeneous type to a `Schema` instance.

    :param arg: Argument to resolve to a Schema. Must be one of:
                A primitive Python type (str, int, bool, float)
                A subclass of `UserObject`
                An instance of `Schema`.
    :return:    A `Schema` instance corresponding to the supplied argument.
    """
    if inspect.isclass(arg):
        if issubclass(arg, DOMElement):
            return arg.__json_schema__()
        if issubclass(arg, Enum):
            return SchemaEnum(arg)
        else:
            return SchemaPrimitive(arg)
    elif isinstance(arg, Schema):
        return arg
    else:
        raise TypeError(f"Unexpected object type: {type(arg)}")



from typing import Any, Dict, List, Tuple, Type
from enum import Enum

from .Schema import Schema


class SchemaEnum(Schema):
    """
    A schema requiring a one of a set of enumerated values.
    """

    enum: Type[Enum] = None

    def __init__(
            self,
            enum: Type[Enum]
    ) -> None:
        self.enum = enum

    def __call__(
            self,
            value: Any,
            dom_info: Tuple = None
    ) -> Any:
        validated_value = super().__call__(value)
        matched_enum_members = [
            member
            for member in self.enum.__members__.values()
            if member.value == validated_value
        ]
        if len(matched_enum_members) == 0:
            raise ValueError(f"Cannot find a member in {self.enum} with value {value}")
        if len(matched_enum_members) > 1:
            raise ValueError(f"Multiple ambiguous members in {self.enum} with value {value}")
        return matched_enum_members[0]

    @property
    def allowed_values(self) -> List[Any]:
        return [
            member.value
            for member in self.enum.__members__.values()
        ]

    @property
    def jsonschema_definition(self) -> Dict[str, Any]:
        return {
            "enum": self.allowed_values
        }
